/*
Copyright (C) 1998, 1999, 2000 Wabasoft

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA. 
*/

/*
Copyright (C) 2000 Kenta Cho(cs8k-cyu@asahi-net.or.jp)
 Waba for Ruputer
 http://www.asashi-net.or.jp/~cs8k-cyu/
*/
 
/*

If you're looking here, you've probably already looked at (or worked on) the
nm<platform>_a.c file.

This file is the second step in porting the VM. In it, you need to place the
main application loop and class loader. If you get through this, you'll have
a basic VM running and from there you can implement the native functions
so you can actually see something on the screen.

The WabaVM, like other applications, needs some type of main application loop
where the OS sends it key clicks, mouse presses, etc. The VM also needs to
load classes from memory or disk or wherever.

Different platforms have different main application loops. For example, under
Win32, a program has a WinMain() function. Under PalmOS, applications have
a PilotMain() function.

You'll want to implement a main application loop for the platform you are
porting to. Then, when things start up (initialization), you need to:

- parse the launch command and then call VmInit()
- call VmStartApp passing the class name of the program's main window

You'll also need to figure out how programs will launch the VM. The VM
is normally passed parameters telling it what class heap, class name, etc.
to use. It's usually another program that does this but you could do it
by building a waba launcher that draws a tray of icons and when you click
one, it runs the WabaVM with various parameters as an example of something
different. Whatever you do for this is platform-specific.

Just before the program exits, you'll should call VmStopApp(mainWinObj);

That's it for initialization and exit, then you need to do some work to
hook up key presses, mouse presses, etc. to the VM. When a key is pressed,
you'll want to invoke the main window's onEvent() method. Here's an
example of how to do that:

	mainWinObj = globalMainWin;
	if (!mainWinObj)
		return;
	vclass = WOBJ_class(mainWinObj); // get runtime class
	method = getMethod(vclass, createUtfString("_postEvent"),
		createUtfString("(IIIIII)V"), &vclass);
	if (method != NULL)
		{
		params[0].obj = mainWinObj;
		params[1].intValue = type; // type
		params[2].intValue = key; // key
		params[3].intValue = 0; // x
		params[4].intValue = 0; // y
		params[5].intValue = 0; // modifiers
		params[6].intValue = 0; // timeStamp
		executeMethod(vclass, method, params, 7);
		}

This gets the runtime class of the main window object, gets a
reference to the _postEvent method, fills in an array of parameters
to pass and then calls executeMethod() to execute the method.

There is also a quickbind version of this which is quicker than the
above. You can look at nmpalm_b.c or nmwin32_b.c to get an idea how
that works (look for the #ifdef QUICKBIND) but you should start
without using the quickbind method.

You'll want to hook up mouse or pen click, key presses, calls from
the OS to repaint the window (which should call the main window's
_doPaint routine).

Overall, you need to make sure the following main window functions
get called:

_onTimerTick()
_postEvent() for key presses and mouse/pen clicks and moves

You'll need to map special keys (escape, etc.) if the device has
them to suitable numbers for waba. To do this, you can see how the
other native implementations do key mapping (a big switch statement).

The _onTimerTick() one can be tricky for devices that don't have
a built in timer (like PalmOS). When using one of those devices,
copy the code from the nmpalm_b.c to deal with timers. Basically,
this consists of getting the current time and looking for an event
up until the time the next timer period expires.

The VM only needs a single timer, the waba core classes have the
code to support multiple timers even though the underlying system
only supports a single one.

The last code you need to implement is:

nativeLoadClass()

This is the function that load a class from memory, disk or whatever.
It looks like this:

static uchar *nativeLoadClass(UtfString className, uint32 *size)

and it is passed a class name (not zero terminated, the UtfString
contains the len in:

className.len

and it returns a pointer to the class in memory after it is loaded
and fills in the *size variable with the size of the class in bytes.

The classes loaded are classes that are compiled by a java compiler
or other compiler that can generate code in the subset of java
bytecode that waba supports.

*/

#define MAX_CLASS 32
static uint32 classRecordOfs[MAX_CLASS];
static uint32 classRecordSize[MAX_CLASS];
static int numClassRecords;

static uchar *classesMap;
static WObject globalMainWin = 0;

// search class bytes from memory map
static uchar *nativeLoadClass(UtfString className, uint32 *size) {
	int i;
	char cf[className.len + 7];
	int nameLen;

	//char *loadMsg[2] = {"load", "              "};
	//xstrncpy(loadMsg[1], className.str, 13);
	//dispMess(loadMsg, 2);

	// make class file path
	xstrncpy(cf, className.str, className.len);
	xstrncpy(&cf[className.len], ".class", 6);

	for ( i=0 ; i<numClassRecords ; i++ ) {
		if ( xstrncmp(
		 cf, (char *)(classesMap+classRecordOfs[i]+2), className.len+6) == 0 ) {
			nameLen = getUInt16(classesMap+classRecordOfs[i]);
			*size = classRecordSize[i]-nameLen-2;
			return (classesMap+classRecordOfs[i]+2+nameLen);
		}
	}
	return NULL;
}

// when load file failed
static void handleDosStatus(int dosSt) {
	char *clErrMsg[2] = {"load wrp file","failed."};

	if ( dosSt == -2 ) {
		dispVOLerr();
		dos_exit();
	} else if ( dosSt == -1 ) {
		dispMess((char **)clErrMsg, 2);
		dos_exit();
	}
}

// load classes wrp file into memory
static void loadWrpWarpFile() {
	int i;
	int fp;
	int len;
	int dosSt;

	char filename[80];
	strcpy(filename,"CLASSES.WRP");
	
	// load wrp warp file(class files)	
	fp = dos_fopen(filename, 0);
	handleDosStatus(fp);
	len = dos_fseek(fp, 2, 0);
	classesMap = xmalloc(len);
	dos_fseek(fp, 0, 0);
	dosSt = dos_fread(fp, (char*)classesMap, (int)len);
	dos_fclose(fp);
	handleDosStatus(dosSt);

	numClassRecords = getUInt32(classesMap+4);
	for ( i=0 ; i<numClassRecords ; i++ ) {
		classRecordOfs[i] = getUInt32(classesMap+8+4*i);
		classRecordSize[i] = getUInt32(classesMap+12+4*i)-classRecordOfs[i];
	}
}

// print VM error message
static void printError() {
	char *erMsg[3] = {"Error:", "              ", "             "};
	char *msg;

	xstrncpy(erMsg[1], errorMessages[vmStatus.errNum-1], 13);
	msg = vmStatus.arg1;
	if ( msg[0] ) {
		xstrncpy(erMsg[2], msg, 12);
	}	
	dispMess(erMsg, 3);
}

// call onPaint method in MainWindow
static void drawMainWin() {
	WObject mainWinObj;
	WClass *vclass;
	WClassMethod *method;
	Var params[3];

	mainWinObj = globalMainWin;
	if (!mainWinObj)
		return;
	vclass = WOBJ_class(mainWinObj); // get runtime class
	method = getMethod(vclass, createUtfString("_doPaint"),
		createUtfString("()V"), &vclass);
	if (method == NULL)
		return;
	params[0].obj = mainWinObj;
	executeMethod(vclass, method, params, 1);
}

#ifdef QUICKBIND
static int32 postEventMethodMapNum = -1;
static int32 onTimerTickMethodMapNum = -1;
#endif

// count tick (1/64 sec)
int32 tick = 0;

static void timer64Func() {
	tick++;
}

static int32 getTimeStamp() {
	return tick*1000/64;
}

int32 globalTimerStart;
int32 globalTimerInterval = 0;

// check timer interval
static void timerCheck() {
	WObject mainWinObj;
	WClass *vclass;
	WClassMethod *method;
	Var params[1];
	int32 now ;
	int32 diff;
	int32 timeout;

	mainWinObj = globalMainWin;
	if (!mainWinObj) return;

	now = getTimeStamp();
	diff = now - globalTimerStart;
	//if ( diff < 0 ) {
	//	diff += (1L << 30);
	//}
	timeout = globalTimerInterval - diff;
	if ( timeout > 0 ) return;

	vclass = WOBJ_class(mainWinObj); // get runtime class
#ifdef QUICKBIND
	method = getMethodByMapNum(vclass, &vclass,
		(uint16)onTimerTickMethodMapNum);
#else
	method = getMethod(vclass, createUtfString("_onTimerTick"),
		createUtfString("()V"), &vclass);
#endif
	if (method != NULL) {
		params[0].obj = mainWinObj;
		executeMethod(vclass, method, params, 1);
	}
	globalTimerStart = now;
}

// call postEvent method
static void postKeyPress(int32 keyCode) {
	WObject mainWinObj;
	WClass *vclass;
	WClassMethod *method;
	Var params[7];
	mainWinObj = globalMainWin;
	vclass = WOBJ_class(mainWinObj); // get runtime class
#ifdef QUICKBIND
	method = getMethodByMapNum(vclass, &vclass,
		(uint16)postEventMethodMapNum);
#else
	method = getMethod(vclass, createUtfString("_postEvent"),
		createUtfString("(IIIIII)V"), &vclass);
#endif
	if (method != NULL) {
		params[0].obj = mainWinObj;
		params[1].intValue = 100; // type
		params[2].intValue = keyCode; // key
		params[3].intValue = 0; // x
		params[4].intValue = 0; // y
		params[5].intValue = 0; // modifiers
		params[6].intValue = 0; // timeStamp
		executeMethod(vclass, method, params, 7);
	}
}

#define KEY_FILER 1
#define KEY_MENU  2
#define KEY_ENTER 8
#define KEY_LEFT  16
#define KEY_UP    32
#define KEY_RIGHT 64
#define KEY_DOWN  128

static void handleKeyPress(int key) {
	if ( (key&KEY_MENU) == KEY_MENU ) {
		postKeyPress(75014);
	}
	if ( (key&KEY_ENTER) == KEY_ENTER ) {
		postKeyPress(75009);
	}
	if ( (key&KEY_LEFT) == KEY_LEFT ) {
		postKeyPress(75006);
	}
	if ( (key&KEY_UP) == KEY_UP ) {
		postKeyPress(75004);
	}
	if ( (key&KEY_RIGHT) == KEY_RIGHT ) {
		postKeyPress(75007);
	}
	if ( (key&KEY_DOWN) == KEY_DOWN ) {
		postKeyPress(75005);
	}
}

int mainWinExit = 0;

// handle timer & key event
static void startEventLoop() {
	int key;
	//int prvKey = 0;
	int *timerID;
	timerID = bi_tmset(2, 255, 1, timer64Func);
	do {
		/*int pk;
		pk = key = bi_scanbtn();
		pk = key = bi_scanbtn();
		key &= ~prvKey;
		prvKey = pk;
		handleKeyPress(key);*/
		int pushd;
		pushd = bi_sensbtn(&key);
		if ( pushd ) {
			handleKeyPress(key);
			bi_clrbtn();
		}
		if ( globalTimerInterval >= 0 ) {
			timerCheck();
		}
	} while (
	 (key&KEY_FILER) != KEY_FILER
	  && vmStatus.errNum == 0
	  && !mainWinExit );
	bi_tmdel(timerID);
}

// VM init & start
static int RuputerMain(void) {
	unsigned long vmStackSize, nmStackSize, classHeapSize, objectHeapSize;
	char *className = "MainClass";
	WObject mainWinObj;

	// defaults
	vmStackSize = 400;
	nmStackSize = 100;
	classHeapSize = 4800; // wrp file size * .2
	objectHeapSize = 1600; // class heap size / 3
	
	dispMessBar("VmInit");
	loadWrpWarpFile();
	VmInit(vmStackSize, nmStackSize, classHeapSize, objectHeapSize);
	mainWinObj = VmStartApp(className);
#ifdef QUICKBIND
	if (mainWinObj != 0) {
		WClass *vclass;

		// cache method map numbers for commonly called methods
		vclass = WOBJ_class(mainWinObj);
		postEventMethodMapNum = getMethodMapNum(vclass, createUtfString("_postEvent"),
			createUtfString("(IIIIII)V"), SEARCH_ALL);
		onTimerTickMethodMapNum = getMethodMapNum(vclass, createUtfString("_onTimerTick"),
			createUtfString("()V"), SEARCH_ALL);
		if (postEventMethodMapNum == -1 || onTimerTickMethodMapNum == -1)
			mainWinObj = 0;
	}
#endif
	endMessBar();
	cls(4);
	if ( vmStatus.errNum == 0 ) {
		drawMainWin();
	}
	startEventLoop();
	if ( vmStatus.errNum != 0 ) {
		printError();
	}
	VmStopApp(mainWinObj);
}
